package blatt8vorlage;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;

public class Node {
	
	/**
	 * This constant is used for edgeLabelEnd and indicates the end of the respective string.
	 */
	public static int INFINITY_LABEL = Integer.MAX_VALUE-1; // "minus one" is important, because sometimes we use expressions like "edgeLabelEnd + 1"

	private int edgeLabelBegin; // inclusive
	private int edgeLabelEnd; // inclusive
	private Node parent;
	private Node suffixLink;
	private HashMap<Character, Node> children;
	private boolean isSuperRoot, isRoot;

	public Node(Node parent, int edgeLabelBegin, int edgeLabelEnd, boolean isSuperRoot, boolean isRoot) {
		this.parent = parent;
		this.edgeLabelBegin = edgeLabelBegin;
		this.edgeLabelEnd = edgeLabelEnd;
		this.isSuperRoot = isSuperRoot;
		this.isRoot = isRoot;
		this.children = new HashMap<Character, Node>();
		this.suffixLink = null;
	}

	public Node(Node parent, int edgeLabelBegin, int edgeLabelEnd) {
		this(parent, edgeLabelBegin, edgeLabelEnd, false, false);
	}
	
	public Node(Node parent, int edgeLabelBegin) {
		this(parent, edgeLabelBegin, INFINITY_LABEL, false, false);
	}
	

	public Node getParent() {
		return parent;
	}

	public void setParent(Node newParent) {
		parent = newParent;
	}
	
	/**
	 * Adds a new child to this node. The hash map key is determined by the child's edge label.
	 * If there is already a child with the same first character, then the old child is overwritten
	 * by the new one.
	 */
	public void addChild(Node child, String text) {
		children.put(text.charAt(child.edgeLabelBegin), child);
	}
	
	/**
	 * Adds the root to the superRoot. For the superRoot's child we use the key "null". 
	 */
	public void addRoot(Node root) {
		children.put(null, root);
	}

	/**
	 * Checks if there is a child which label begins with "c". For the superRoot we always return true.
	 */
	public boolean containsChild(char c) {
		return isSuperRoot || children.containsKey(c);
	}

	/**
	 * @return the child which edge label begins with "c". For the superRoot we always return the root. If there is no child which label begins with "c", null is returned.
	 */
	public Node getChild(char c) {
		return isSuperRoot ? children.get(null) : children.get(c);
	}
	
	public HashMap<Character, Node> getChildren() {
		return children;
	}
	
	public int getEdgeLabelBegin() {
		return edgeLabelBegin;
	}
	
	public int getEdgeLabelEnd() {
		return edgeLabelEnd;
	}
	
	public void setEdgeLabelBegin(int edgeLabelBegin) {
		this.edgeLabelBegin = edgeLabelBegin;
	}
	
	public void setEdgeLabelEnd(int edgeLabelEnd) {
		this.edgeLabelEnd = edgeLabelEnd;
	}
	
	public int getEdgeLength(String text) {
		// take the minimum, because edgeLabelEnd can be equal to INFINITY_LABEL.
		return Math.min(text.length(), edgeLabelEnd) - edgeLabelBegin + 1;
	}
	

	public Node getSuffixLink() {
		return suffixLink;
	}

	public void setSuffixLink(Node suffixLink) {
		this.suffixLink = suffixLink;
	}
	
	
	public boolean isLeaf() {
		return children.isEmpty();
	}

	public boolean isSuperRoot() {
		return isSuperRoot;
	}

	public boolean isRoot() {
		return isRoot;
	}
	
	/**
	 * @return the represented substring of the text
	 */
	public String getEdgeString(String text) {
		return getEdgeString(text, text.length());
	}
	
	/**
	 * @return the represented substring of text[0...length-1]
	 */
	public String getEdgeString(String text, int length) {
		return text.substring(edgeLabelBegin, Math.min(length, edgeLabelEnd+1));
	}
	
	/**
	 * @return the represented substring of the text respectively "root" or "superRoot".
	 */
	public String getNodeString(String text) {
		return isRoot ? "root" : isSuperRoot ? "superRoot" : getEdgeString(text);
	}
	
	/**
	 * @return the newick string of the node
	 */
	public String toNewick(String text) {
		return toNewick(text, text.length());
	}

	/**
	 * Returns the newick string of the node using the first "length" characters of the text.
	 * This method can be useful for debugging messages while building the suffix tree.
	 */
	public String toNewick(String text, int length) {
		if (getChildren().isEmpty() && isRoot)
			return "()";
		else if (getChildren().isEmpty()) {
			return getEdgeString(text, length);
		} else {
			String s = "(";
			boolean first = true;
			for (Node c : sortChildren(text)) {
				s += (first ? "" : ",") + c.toNewick(text, length);
				first = false;
			}
			s += ")";
			return s + (isRoot ? "" : getEdgeString(text));
		}
	}
	
	/**
	 * @return a sorted list of the node's children.
	 */
	private ArrayList<Node> sortChildren(final String text){
		ArrayList<Node> al = new ArrayList<>(children.values());
		Collections.sort(al, new Comparator<Node>() {
			@Override
			public int compare(Node v1, Node v2) {
				char c1 = text.charAt(v1.getEdgeLabelBegin());
				char c2 = text.charAt(v2.getEdgeLabelBegin());
				if(c1 > c1)
					return 1;
				else if(c1 < c2)
					return -1;
				return 0;
			}
		});
		return al;
	}

}
